from __future__ import annotations

import shutil
import subprocess
import tempfile
from pathlib import Path

import typer
from loguru import logger

from .pip import read_requirements_file


def write_requirements_file_from_project_dir(
    project_dir: Path,
    out_path: Path,
    extra_args: list[str] | None = None,
) -> None:
    _check_uv_installed()
    command = [
        "uv",
        "export",
        "--project",
        project_dir,
        "--no-emit-project",
        "--no-dev",
        "--no-hashes",
        "--quiet",
        "--output-file",
        out_path,
    ]
    if extra_args is not None:
        command.extend(extra_args)
    command = [str(arg) for arg in command]

    try:
        subprocess.run(
            command,
            capture_output=True,
            text=True,
            check=True,
        )
    except subprocess.CalledProcessError as e:
        command_str = " ".join(command)
        msg = (
            "Error creating requirements file from uv project."
            f"\nCommand: {command_str}"
            f"\nOutput from uv: {e.stderr.strip()}"
        )
        logger.error(msg)
        raise typer.Exit(1) from e


def get_requirents_from_project_dir(
    project_dir: Path,
    uv_args: list[str] | None = None,
    out_requirements_path: Path | None = None,
) -> list[str]:
    if out_requirements_path is None:
        with tempfile.NamedTemporaryFile(mode="w", suffix=".txt") as f:
            requirements_path = Path(f.name)
            write_requirements_file_from_project_dir(
                project_dir,
                requirements_path,
                extra_args=uv_args,
            )
            requirements = read_requirements_file(requirements_path)
    else:
        write_requirements_file_from_project_dir(
            project_dir,
            out_requirements_path,
            extra_args=uv_args,
        )
        requirements = read_requirements_file(out_requirements_path)
    return requirements


def _check_uv_installed() -> None:
    if shutil.which("uv") is None:
        url = "https://docs.astral.sh/uv/getting-started/installation"
        msg = f"uv not found. Please install it: {url}"
        raise RuntimeError(msg)



from pathlib import Path


def read_requirements_file(requirements_file: Path) -> list[str]:
    """Read a requirements file and return a list of requirements.

    Removes comments and empty lines.

    Args:
        requirements_file: Path to the requirements file.

    """
    with requirements_file.open() as f:
        lines = f.readlines()
    requirements = []
    for line in lines:
        stripped = line.strip()
        if not stripped or stripped.startswith("#"):
            continue
        requirements.append(stripped)
    return requirements



from __future__ import annotations

from typing import TYPE_CHECKING

import yaml

from .pip import read_requirements_file
from .python import get_python_version_from_project_dir
from .python import is_valid_python_version
from .uv import get_requirents_from_project_dir

if TYPE_CHECKING:
    from pathlib import Path

TypePipEnv = dict[str, list[str]]
TypeCondaDependency = str | TypePipEnv
TypeChannels = list[str]
TypeCondaEnv = dict[str, str | TypeChannels | list[TypeCondaDependency]]


def make_conda_env_from_dependencies(
    name: str,
    python_version: str,
    channels: list[str] | None = None,
    conda_dependencies: list[str] | None = None,
    pip_dependencies: list[str] | None = None,
    out_path: Path | None = None,
) -> TypeCondaEnv:
    if not is_valid_python_version(python_version):
        msg = f'Invalid Python version: "{python_version}"'
        raise ValueError(msg)
    env_dict: TypeCondaEnv = {
        "name": name,
    }
    if channels:
        env_dict["channels"] = channels
    if conda_dependencies or pip_dependencies:
        dependencies_list: list[TypeCondaDependency] = [
            f"python={python_version}",
        ]
        if conda_dependencies:
            dependencies_list.extend(conda_dependencies)
        if pip_dependencies:
            dependencies_list.append("pip")
            dependencies_list.append({"pip": pip_dependencies})

        env_dict["dependencies"] = dependencies_list

    if out_path is not None:
        env_to_file(env_dict, out_path)

    return env_dict


def env_to_str(env: TypeCondaEnv | str) -> str:
    if isinstance(env, str):
        env_string = env
    else:
        env_string = yaml.dump(env, sort_keys=False, width=1000)
    return env_string


def env_to_file(env: TypeCondaEnv, out_path: Path) -> None:
    with out_path.open("w") as f:
        f.write(env_to_str(env))


def make_conda_env_from_requirements_file(
    name: str,
    python_version: str,
    requirements_path: Path,
    channels: list[str] | None = None,
    conda_dependencies: list[str] | None = None,
    out_path: Path | None = None,
) -> TypeCondaEnv | str:
    return make_conda_env_from_dependencies(
        name,
        python_version,
        channels=channels,
        conda_dependencies=conda_dependencies,
        pip_dependencies=read_requirements_file(requirements_path),
        out_path=out_path,
    )


def make_conda_env_from_project_dir(
    project_dir: Path,
    name: str | None = None,
    python_version: str | None = None,
    channels: list[str] | None = None,
    conda_dependencies: list[str] | None = None,
    out_path: Path | None = None,
    uv_args: list[str] | None = None,
    requirements_path: Path | None = None,
) -> TypeCondaEnv | str:
    pip_requirements = get_requirents_from_project_dir(
        project_dir,
        uv_args=uv_args,
        out_requirements_path=requirements_path,
    )
    if name is None:
        name = project_dir.name
    if python_version is None:
        python_version = get_python_version_from_project_dir(project_dir)
        if python_version is None:
            msg = (
                "A Python version must be specified either via the "
                "--python-version option or by creating a "
                ".python-version file in the project directory."
            )
            raise ValueError(msg)
    return make_conda_env_from_dependencies(
        name,
        python_version,
        channels=channels,
        conda_dependencies=conda_dependencies,
        pip_dependencies=pip_requirements,
        out_path=out_path,
    )



from pathlib import Path

from packaging.version import InvalidVersion
from packaging.version import Version


def is_valid_python_version(version: str) -> bool:
    """Check if a string is a valid Python version.

    Args:
        version: The version string to validate.

    Returns:
        bool: True if the version is valid, False otherwise.

    Raises:
        InvalidVersion: If the version string is not a valid version.
    """
    try:
        Version(version)
    except InvalidVersion:
        return False
    else:
        return True


def get_python_version_from_project_dir(project_dir: Path) -> str | None:
    """Get the Python version from the project directory.

    This function looks for a file named `.python-version` in the project
    directory and returns the version specified in that file. If the file
    does not exist, it returns None.

    Args:
        project_dir: The project directory.

    Returns:
        str | None: The Python version or None if not found.
    """
    pinned_version_path = project_dir / ".python-version"
    if pinned_version_path.exists():
        return pinned_version_path.read_text().strip()
    else:
        return None



import importlib.metadata

from .conda import make_conda_env_from_dependencies
from .conda import make_conda_env_from_project_dir
from .conda import make_conda_env_from_requirements_file
from .pip import read_requirements_file
from .uv import get_requirents_from_project_dir
from .uv import write_requirements_file_from_project_dir

__all__ = [
    "get_requirents_from_project_dir",
    "make_conda_env_from_dependencies",
    "make_conda_env_from_project_dir",
    "make_conda_env_from_requirements_file",
    "read_requirements_file",
    "write_requirements_file_from_project_dir",
]

__version__ = importlib.metadata.version(__name__)
